/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_ipt.cc,v 1.16 2002/05/13 18:38:20 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "NATCompiler_ipt.hh"

#include "fwcompiler/OSConfigurator.hh"

#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/NAT.hh"
#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/Host.hh"
#include "fwbuilder/Network.hh"

#include <algorithm>
#include <functional>
#include <iomanip.h>
#include <iostream.h>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

string NATCompiler_ipt::myPlatformName() { return "iptables"; }

string NATCompiler_ipt::getNewTmpChainName(NATRule *rule)
{
    std::ostringstream str;

    str << "ntmp" << setw(3) << setfill('0') << chain_no;
    chain_no++;

    chains[str.str()]=true;
    return str.str();
}

string NATCompiler_ipt::getNewChainName(NATRule *rule)
{
    Interface *rule_iface = fw_interfaces[rule->getStr("interface_id")];

    std::ostringstream str;

    if (rule_iface)  str << "I";
    str << "NAT_RULE_" << rule->getPosition();
    if (rule_iface) str << "_" << rule_iface->getName();

    chain_no++;

    chains[str.str()]=true;

    return str.str();
}

string NATCompiler_ipt::_internalPrintNATRule(libfwbuilder::NATRule *rule)
{
    return NATCompiler::_internalPrintNATRule(rule)+
        " "+rule->getStr("interface_id")+
        " "+rule->getStr("ipt_chain")+
        " "+rule->getStr("ipt_target")+
        " (type="+rule->getRuleTypeAsString()+")";
}


int NATCompiler_ipt::prolog()
{
    standard_chains["POSTROUTING"]  =true;
    standard_chains["PREROUTING"]   =true;


    int n=NATCompiler::prolog();

    if ( n>0 ) {

	bool found_ext=false;
	list<FWObject*> l2=fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    if ( iface->isExt() ) found_ext=true;
	}

	if (!found_ext)
	    throw FWException("At least one interface should be marked as external, can not configure NAT");
    }

/* pseudo-host with ip address 127.0.0.1  We'll use it for redirection NAT rules 
    FWObject    *grp;
    loopback_address=Host::cast(dbcopy->create(Host::TYPENAME,false) );
    loopback_address->setStr("address","127.0.0.1");
    grp=dbcopy->getById( dbcopy->std.HostsId , true );
    assert(grp!=NULL);
    grp->add(loopback_address);
*/

    return n;
}

void NATCompiler_ipt::NATRuleType::operator()(NATRule *rule) const
{
    Address  *osrc=rule->getOSrc()->getFirst(true);
    Address  *odst=rule->getODst()->getFirst(true);

    Address  *tsrc=rule->getTSrc()->getFirst(true);
    Address  *tdst=rule->getTDst()->getFirst(true);

    if (   tsrc->isAny() &&   tdst->isAny() ) {
	rule->setRuleType(NATRule::NONAT);
	return;
    }

    if ( ! tsrc->isAny() &&   tdst->isAny() ) {
        if ( Network::isA(tsrc) )
/* 
 * this is Netnat rule ( NETMAP in iptables) 
 * we do some additional sanity checks in VerifyRules, see below
 */
            rule->setRuleType(NATRule::SNetnat);
        else 
            rule->setRuleType(NATRule::SNAT);
        return;
    }

    if (   tsrc->isAny() && ! tdst->isAny() ) {
        if ( Network::isA(tdst) )
/* 
 * this is Netnat rule ( NETMAP in iptables) 
 * we do some additional sanity checks in VerifyRules, see below
 */
            rule->setRuleType(NATRule::DNetnat);
        else {
            if ( tdst==getCompiler()->fw ) rule->setRuleType(NATRule::Redirect);
            else                           rule->setRuleType(NATRule::DNAT);
        }
	return;
    }


    throw FWException("Unsupported NAT rule: "+rule->getStr("rule_label"));
}

void NATCompiler_ipt::VerifyRules::operator()(NATRule *rule) const
{
    RuleElementOSrc  *osrc=rule->getOSrc();  assert(osrc);
    RuleElementODst  *odst=rule->getODst();  assert(odst);
    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);

    RuleElementTSrc  *tsrc=rule->getTSrc();  assert(tsrc);
    RuleElementTDst  *tdst=rule->getTDst();  assert(tdst);
    RuleElementTSrv  *tsrv=rule->getTSrv();  assert(tsrv);

    if (tsrc->getNeg())
	throw FWException("Can not use negation in translated source. Rule "+rule->getStr("rule_label"));

    if (tdst->getNeg())
	throw FWException("Can not use negation in translated destination. Rule "+rule->getStr("rule_label"));

    if (tsrv->getNeg())
	throw FWException("Can not use negation in translated service. Rule "+rule->getStr("rule_label"));

    if (tsrv->size()!=1) 
	throw FWException("Translated service should be 'Original' or should contain single object. Rule: "+rule->getStr("rule_label"));

    if ( Group::cast( tsrv->getFirst(true) )!=NULL)
	throw FWException("Can not use group in translated service. Rule "+rule->getStr("rule_label"));



    if (rule->getRuleType()==NATRule::DNAT) {

        if ( odst->isAny() )
            throw FWException("Original destination can not be 'any' in DNAT rule. Rule "+rule->getStr("rule_label"));
        if ( odst->size()!=1)
            throw FWException("There should be no more than one object in original destination in the rule "+rule->getStr("rule_label"));

        Address* o2=odst->getFirst(true);
        if ( ! Host::isA(o2)  && ! Firewall::isA(o2) )
            throw FWException("Original destination must be host or firewall for this type of NAT rule. Rule "+rule->getStr("rule_label"));
/*
 *  check if we have specific IP address to use in odst 
 */
        bool find_ext_static_address=false;
        FWObjectTypedChildIterator j=o2->findByType(Interface::TYPENAME);
        for ( ; j!=j.end(); ++j ) {
            Interface *iface=Interface::cast(*j);
            if ( ! iface->isLoopback() && iface->isExt() && ! iface->isDyn() ) 
                find_ext_static_address=true;
        }
        if ( ! find_ext_static_address && Host::isA(o2) ) 
            find_ext_static_address=true;

        if ( ! find_ext_static_address ) {
            if ( Firewall::isA(o2) )
                throw FWException("Could not find appropriate address for original destination: external interface of the firewall has dynamic address. Rule "+rule->getStr("rule_label"));
            else
                throw FWException("Could not find appropriate address for original destination. Rule "+rule->getStr("rule_label"));
        }
    }

    if (rule->getRuleType()==NATRule::SNAT ) {
        if ( tsrc->size()!=1)
            throw FWException("There should be no more than one object in translated source in the rule "+rule->getStr("rule_label"));

        Address* o1=tsrc->getFirst(true);
        if ( ! tsrc->isAny() && Network::cast(o1)!=NULL)
            throw FWException("Can not use network object in translated source. Rule "+rule->getStr("rule_label"));
    }


    if (rule->getRuleType()==NATRule::SNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(osrc->getFirst(true));
        Network *a2=Network::cast(tsrc->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException("Original and translated source should both be networks of the same size . Rule "+rule->getStr("rule_label"));
    }

    if (rule->getRuleType()==NATRule::DNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(odst->getFirst(true));
        Network *a2=Network::cast(tdst->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException("Original and translated destination should both be networks of the same size . Rule "+rule->getStr("rule_label"));
    }


}

void NATCompiler_ipt::splitOnOSrv::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);
    if (osrv->size()!=1) {

	for(list<FWObject*>::iterator i=osrv->begin(); i!=osrv->end(); ++i) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) 
		o=FWReference::cast(o)->getPointer();
	    Service *s=Service::cast( o );
	    assert(s);

	    NATRule *r= NATRule::cast(getCompiler()->
				      dbcopy->create(NATRule::TYPENAME,true) );
	    getCompiler()->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementOSrv *nosrv=r->getOSrv();
	    nosrv->clearChildren();

	    nosrv->addRef( s );

	    vr->push_back( r );
	}
    } else
	vr->push_back(rule);
}

void NATCompiler_ipt::fillTranslatedSrv::operator()(NATRule *rule) const
{
    Service  *osrv_o=rule->getOSrv()->getFirst(true);
    Service  *tsrv_o=rule->getTSrv()->getFirst(true);

    if ( ! osrv_o->isAny()  && tsrv_o->isAny() ) {
	RuleElementTSrv  *tsrv=rule->getTSrv();
	tsrv->addRef(osrv_o);
    }
}

void NATCompiler_ipt::addVirtualAddress::operator()(NATRule *rule) const
{
    Address *a=NULL;
    if (rule->getRuleType()==NATRule::SNAT) a=rule->getTSrc()->getFirst(true);
    else
	if (rule->getRuleType()==NATRule::DNAT) a=rule->getODst()->getFirst(true);
	else  return;
    assert(a!=NULL);

    if (a->getId()!=getCompiler()->fw->getId() ) {

	list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    if ( a->getAddress() == iface->getAddress() ) return;
	}
	getCompiler()->osconfigurator->addVirtualAddressForNAT( a->getAddress() );
    }
}

void NATCompiler_ipt::splitRuleIfRuleElementIsDynamicInterface(NATRule *rule,
                                                               std::vector<Rule*> *vr,
                                                               const std::string &re_type)
{
    RuleElement    *re=RuleElement::cast(rule->getFirstByType(re_type));
    int nre=re->size();

    vector<FWObject*> cl;

    for(list<FWObject*>::iterator i=re->begin(); nre>1 && i!=re->end(); ++i) {
        FWObject *o= *i;
	FWObject *obj = NULL;
        if (FWReference::cast(o)!=NULL) obj=objcache[o->getStr("ref")];
        Interface *iface=Interface::cast(obj);
        if (iface!=NULL && iface->isDyn()) {

	    cl.push_back(o);   // can not remove right now because remove invalidates iterator
            nre--;

	    NATRule  *new_rule= NATRule::cast(dbcopy->create(NATRule::TYPENAME,true) );
	    temp_ruleset->add(new_rule);
	    new_rule->duplicate(rule);
            RuleElement *new_re=RuleElement::cast(new_rule->getFirstByType(re_type));
	    new_re->clearChildren();
	    new_re->setAnyElement();
	    new_re->addRef( iface );
	    vr->push_back(new_rule);
        }
    }
    if (!cl.empty()) {
        for (vector<FWObject*>::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  
            re->remove( (*i1) );
    }

    vr->push_back(rule);
}

void NATCompiler_ipt::splitOnDynamicInterfaceInTSrc::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    getCompiler()->splitRuleIfRuleElementIsDynamicInterface(rule,vr,RuleElementTSrc::TYPENAME);
}

void NATCompiler_ipt::splitOnDynamicInterfaceInODst::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    getCompiler()->splitRuleIfRuleElementIsDynamicInterface(rule,vr,RuleElementODst::TYPENAME);
}

/*
 *  I assume that there is always only one object in ODst, TSrc and TDst
 *  rule elements. This should have been assured by inspector VerifyRules
 */
void NATCompiler_ipt::ReplaceFirewallObjectsODst::operator()(NATRule *rule) const
{
    list<FWObject*> cl;
    RuleElementODst *rel;
    Address     *obj=NULL;

    switch (rule->getRuleType()) {

    case NATRule::Masq:
//    case NATRule::Redirect: 
	return;
    default:
	rel=rule->getODst();      assert(rel);
	obj=rel->getFirst(true);  assert(obj!=NULL);

	if (obj->getId()==getCompiler()->fw_id ) 
	{

	    list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
		Interface *interface_=Interface::cast(*i);
/*
 * Right now I build DNAT and Redirect rules only for external interface.
 * This actually implies some sort of assymmetric firewall, which does
 * not have to be like this.
 * 
 * TODO: add platform-specific fw option "Generate NAT rules for all
 * interfaces" and use here all interfaces , not only external ones.
 * Do this for v1.0.4.  Modify verification code in verifyRules
 * accordingly. The same applies to ReplaceFirewallObjectsTSrc
 */
		if (! interface_->isLoopback() && 
                      interface_->isExt() &&
                    ! interface_->isDyn() ) cl.push_back(interface_);
	    }
	    if ( ! cl.empty() ) {
		while (rel->size()) 
		    rel->remove( rel->front()  );
		
		for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
		{
		    rel->addRef( *i1 );
		}
	    }
	}
    }
}

void NATCompiler_ipt::ReplaceFirewallObjectsTSrc::operator()(NATRule *rule) const
{
    list<FWObject*> cl;
    RuleElementTSrc *rel;
    Address     *obj=NULL;

    switch (rule->getRuleType()) {

    case NATRule::Masq:
    case NATRule::Redirect: 
	return;
    default:
	rel=rule->getTSrc();      assert(rel);
	obj=rel->getFirst(true);  assert(obj!=NULL);

	if (obj->getId()==getCompiler()->fw_id ) 
	{

	    list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
		Interface *interface_=Interface::cast(*i);
/*
 *  see comment above in ReplaceFirewallObjectsODst
 */
		if (! interface_->isLoopback() &&
                      interface_->isExt() && 
                    ! interface_->isDyn() ) cl.push_back(interface_);
	    }
	    if ( ! cl.empty() ) {
		while (rel->size()) 
		    rel->remove( rel->front()  );
		
		for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
		{
		    rel->addRef( *i1 );
		}
	    }
	}
    }
}

void NATCompiler_ipt::dynamicInterfaceInODst::operator()(NATRule *rule) const
{
    RuleElementODst  *odstrel=rule->getODst();  assert(odstrel);
    Address          *odst   =odstrel->getFirst(true);
    if ( ! odstrel->isAny() && 
	 Interface::cast(odst)!=NULL &&
	 Interface::cast(odst)->isDyn() ) {

	odstrel->clearChildren();
	odstrel->setAnyElement();
    }
}

void NATCompiler_ipt::dynamicInterfaceInTSrc::operator()(NATRule *rule) const
{
    Address  *tsrc=rule->getTSrc()->getFirst(true);

    if (rule->getRuleType()==NATRule::SNAT && 
        Interface::cast(tsrc)!=NULL &&
	Interface::cast(tsrc)->isDyn()) {

            rule->setRuleType(NATRule::Masq);

            if ( rule->getStr("ipt_target")=="" || rule->getStr("ipt_target")=="SNAT" )
                rule->setStr("ipt_target","MASQ");
    }
}

/**
 * unlike standard inspector addressRanges in the base class NATCompiler,
 * this one does not expand address ranges in TSrc and TDst because
 * iptables supports ranges in those rule elements
 */
void NATCompiler_ipt::addressRanges::operator()(NATRule *rule) const
{
    RuleElement *rel;

    rel=rule->getOSrc();    assert(rel);
    getCompiler()->_expandAddressRanges(rule,rel);
    rel=rule->getODst();    assert(rel);
    getCompiler()->_expandAddressRanges(rule,rel);
}



void NATCompiler_ipt::splitServices::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrv *srv=rule->getOSrv();

    if (srv->size()==1) {
	vr->push_back(rule);
	return;
    }

    map<int, list<Service*> > services;

 loop:
    for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) {
	FWObject *o= *i;
//	if (FWReference::cast(o)!=NULL) o=FWReference::cast(o)->getPointer();
	if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];

	Service *s=Service::cast( o );
	assert(s);

	int proto=s->getProtocolNumber();
	services[proto].push_back(s);
    }

    for (map<int, list<Service*> >::iterator i=services.begin(); i!=services.end(); i++) {
	list<Service*> &sl=(*i).second;

	NATRule *r= NATRule::cast(
	    getCompiler()->dbcopy->create(NATRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	RuleElementOSrv *nsrv=r->getOSrv();
	nsrv->clearChildren();

	for (list<Service*>::iterator j=sl.begin(); j!=sl.end(); j++) {
	    nsrv->addRef( (*j) );
	}

	vr->push_back(r);

    }
}

void NATCompiler_ipt::prepareForMultiport::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrv *rel= rule->getOSrv();
    Service         *srv= rel->getFirst(true);

    if (rel->size()==1) {
	vr->push_back(rule);
	return;
    }

/*
 * processor splitServices should have been called eariler, so now all
 * services in Srv are of the same type
 */
    if (IPService::isA(srv) || ICMPService::isA(srv)) {
/* multiport does  not support ip and icmp services, split the rule */
	for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) {
	    FWObject *o= *i;
//	    if (FWReference::cast(o)!=NULL) o=FWReference::cast(o)->getPointer();
	    if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
	    
	    Service *s=Service::cast( o );
	    assert(s);

	    NATRule *r= NATRule::cast(
		getCompiler()->dbcopy->create(NATRule::TYPENAME,false) );
	    getCompiler()->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementOSrv *nsrv=r->getOSrv();
	    nsrv->clearChildren();
	    nsrv->addRef( s );

	    vr->push_back(r);
	}
	return;
    }

    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	rule->setBool("ipt_multiport",true);

/* TODO: make sure we have no more than 15 ports */

	vr->push_back(rule);
    }

}

void NATCompiler_ipt::doOSrcNegation::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrc *osrcrel=rule->getOSrc();

/*   ! A  B  C    */

    if (osrcrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	osrcrel->setNeg(false);
/*
 *   negation in OSrc :
 *
 *  CHAIN     !A    B    C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----      any   B    C    SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN   A   any  any   RETURN     RETURN
 * TMP_CHAIN  any  any  any   SNAT/DNAT  ---------
 */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   A  B  any  RETURN  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any B  any  ACTION  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}

void NATCompiler_ipt::doODstNegation::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementODst *odstrel=rule->getODst();

/*   ! A  B  C    */

    if (odstrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	odstrel->setNeg(false);
/*
 *   negation in Odst :
 *
 *  CHAIN      A    !B   C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----       A   any   C    SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN  any   B   any   RETURN     RETURN
 * TMP_CHAIN  any  any  any   SNAT/DNAT  ---------
 */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getODst();   ndst->clearChildren();   ndst->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   any  B  any  RETURN  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any  any  any  ACTION  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}

void NATCompiler_ipt::doOSrvNegation::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrv *osrvrel=rule->getOSrv();

/*   A  B ! C    */

    if (osrvrel->getNeg()) {
	NATRule     *r;
	RuleElementOSrc *nsrc;
	RuleElementODst *ndst;
	RuleElementOSrv *nsrv;

	RuleElementTSrc *ntsrc;
	RuleElementTDst *ntdst;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	osrvrel->setNeg(false);
/*
 *   negation in OSrv :
 *
 *  CHAIN      A    B   !C    RULE_TYPE  TARGET
 *-----------------------------------------------
 * -----       A    B   any   SNAT/DNAT  TMP_CHAIN  
 * TMP_CHAIN  any  any   C    RETURN     RETURN
 * TMP_CHAIN  any  any  any   SNAT/DNAT  ---------
 */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
//	ntsrc=r->getTSrc();  ntsrc->clearChildren();  ntsrc->setAnyElement();
//	ntdst=r->getTDst();  ntdst->clearChildren();  ntdst->setAnyElement();
//	r->setRuleType(NATRule::Continue);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   A  B  any  RETURN  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setRuleType(NATRule::Return);
	r->setStr("ipt_target","RETURN");
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any B  any  ACTION  */
	r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getOSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getODst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getOSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}


void NATCompiler_ipt::decideOnChain::operator()(NATRule *rule) const
{
    if ( ! rule->getStr("ipt_chain").empty() ) return; // already defined

    switch (rule->getRuleType()) {
    case NATRule::SNAT:     rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::SNetnat:  rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::Masq:     rule->setStr("ipt_chain","POSTROUTING"); break;
    case NATRule::DNAT:     rule->setStr("ipt_chain","PREROUTING");  break;
    case NATRule::DNetnat:  rule->setStr("ipt_chain","PREROUTING");  break;
    case NATRule::Redirect: rule->setStr("ipt_chain","PREROUTING");  break;
    }
}


void NATCompiler_ipt::decideOnTarget::operator()(NATRule *rule) const
{
    if ( ! rule->getStr("ipt_target").empty() ) return; // already defined

    switch (rule->getRuleType()) {
    case NATRule::NONAT:    rule->setStr("ipt_target","ACCEPT");   break;
    case NATRule::SNAT:     rule->setStr("ipt_target","SNAT");     break;
    case NATRule::SNetnat:  rule->setStr("ipt_target","NETNAP");   break;
    case NATRule::DNAT:     rule->setStr("ipt_target","DNAT");     break;
    case NATRule::DNetnat:  rule->setStr("ipt_target","NETNAP");   break;
    case NATRule::Masq:     rule->setStr("ipt_target","MASQ");     break;
    case NATRule::Redirect: rule->setStr("ipt_target","REDIRECT"); break;
    case NATRule::Return:   rule->setStr("ipt_target","RETURN");   break;
    }
}

void NATCompiler_ipt::AssignInterface::operator()(NATRule *rule ) const
{
    Address *a=NULL;

    switch (rule->getRuleType()) {

    case NATRule::SNAT:
    case NATRule::Masq:
	a=rule->getTSrc()->getFirst(true);
	break;

    case NATRule::DNAT:
    case NATRule::Redirect:
	a=rule->getODst()->getFirst(true);
	break;

    default: 
	return;
    }

    assert(a!=NULL);

    list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	Interface *iface=Interface::cast(*i);
	assert(iface);

        if (iface->getId()==a->getId()) {
	    rule->setStr("interface_id", iface->getId() );
	    return;
        }

        if ( ! iface->isDyn()) {
/* this is the network interface is connected to */
            IPNetwork n( iface->getAddress() , iface->getNetmask() );
            if ( n.belongs( a->getAddress() ) ) {  // && iface->isExt() ) {
                rule->setStr("interface_id", iface->getId() );
                return;
            }
        }

    }
}


void NATCompiler_ipt::compile()
{
    FWOptions* options=fw->getOptionsObject();
    bool manage_virtual_addr=options->getBool("manage_virtual_addr");


    cout << " Compiling NAT rules for " << fw->getName() << " ..." << endl << flush;

    try {

	Compiler::compile();

        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand groups ...\n" << flush;
	for_each_rule( ExpandGroups(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

//	if (verbose) cout << "split rule on original service ..\n" << flush;
//	for_each_rule( splitOnOSrv(this) );
//       if (debug_rule>=0) for_each_rule( debugRule(this) );

//	if (verbose) cout << "fill translated service ..\n" << flush;
//	for_each_rule( fillTranslatedSrv(this) );
//      if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "determine NAT rule types ..\n" << flush;
	for_each_rule( NATRuleType(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "verify NAT rules ..\n" << flush;
	for_each_rule( VerifyRules(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if ( manage_virtual_addr ) {
	    if (verbose) cout << "add virtual addresses for NAT rules ...\n" << flush;
	    for_each_rule( addVirtualAddress(this) );
            if (debug_rule>=0) for_each_rule( debugRule(this) );
	}

	if (verbose) cout << "decide on chain ...\n" << flush;
	for_each_rule( decideOnChain(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "decide on target ...\n" << flush;
	for_each_rule( decideOnTarget(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "replace references to the firewall ...\n" << flush;
	for_each_rule( ReplaceFirewallObjectsODst(this) );
	for_each_rule( ReplaceFirewallObjectsTSrc(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rule on interfaces with dynamic address ...\n" << flush;
        for_each_rule( splitOnDynamicInterfaceInODst(this) );
        for_each_rule( splitOnDynamicInterfaceInTSrc(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

//	if (verbose) cout << "adjust rules for interfaces with dynamic address ... \n" << flush;
//	for_each_rule( dynamicInterfaceInODst(this) );
//	for_each_rule( dynamicInterfaceInTSrc(this) );
//      if (debug_rule>=0) for_each_rule( debugRule(this) );

        if (verbose) cout << "process negation ...\n" << flush;
        for_each_rule( doOSrcNegation(this) );
        for_each_rule( doODstNegation(this) );
        for_each_rule( doOSrvNegation(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand multiple addresses ...\n" << flush;
	for_each_rule( ExpandMultipleAddresses(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand address range objects  ...\n" << flush;
	for_each_rule( addressRanges(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rules with different protocols ...\n" << flush;
	for_each_rule( splitServices(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "prepare for using multiport ...\n" << flush;
	for_each_rule( prepareForMultiport(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "convert to atomic rules ...\n" << flush;
	for_each_rule( ConvertToAtomicForAddresses(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "assign rules to interfaces ...\n" << flush;
	for_each_rule( AssignInterface(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "adjust rules for interfaces with dynamic address ... \n" << flush;
	for_each_rule( dynamicInterfaceInODst(this) );
	for_each_rule( dynamicInterfaceInTSrc(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

    } catch (FWException &ex) {
	abort(ex.toString());
    }
}


void NATCompiler_ipt::epilog()
{
    try {
	if (verbose) cout << "generate iptables code ... \n" << flush;

	output << "#  temporary chains\n#\n";

	map<string,bool>::iterator i;
	for (i=chains.begin(); i!=chains.end(); ++i)
	    output << "iptables -t nat -N " << (*i).first << endl;

	output << "#\n#\n";

/* print access lists here */
	PrintAtomicRule par(this);

	string current_rule_label="";

	for (list<FWObject*>::iterator k=combined_ruleset->begin(); k!=combined_ruleset->end(); ++k) {
	    NATRule *r = NATRule::cast( *k );
	    string rl=r->getStr("rule_label");
	    if (rl!=current_rule_label) {
		output << "#" << endl;
		output << "#  Rule " << rl << endl;

		string    comm=r->getComment();
		string::size_type c1,c2;
		c1=0;
		while ( (c2=comm.find('\n',c1))!=string::npos ) {
		    output << "# " << comm.substr(c1,c2-c1) << endl;
		    c1=c2+1;
		}
		output << "# " << comm.substr(c1) << endl;
		output << "# " << endl;

		current_rule_label=rl;
	    }
	    par(r);
	}

        output << endl;
        output << endl;

	cout << " NAT rules compiled successfully \n" << flush;

    } catch (FWException ex) {
	abort(ex.toString());
    }
}
