/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_ipt.hh,v 1.6 2002/05/08 06:25:26 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _NATCOMPILER_IPT_HH
#define _NATCOMPILER_IPT_HH

#include "fwcompiler/NATCompiler.hh"

#include <map>

namespace libfwbuilder {
    class Host;
    class IPService;
    class ICMPService;
    class TCPService;
    class UDPService;
};

namespace fwcompiler {


    class NATCompiler_ipt : public NATCompiler {

	protected:

	libfwbuilder::Host  *loopback_address;

	virtual void _printProtocol(libfwbuilder::Service *srv);
	virtual void _printSrcService(libfwbuilder::RuleElementOSrv  *o);
	virtual void _printDstService(libfwbuilder::RuleElementOSrv  *o);

	virtual std::string _printPorts(int rs,int re);
	virtual std::string _printICMP(libfwbuilder::ICMPService *srv);
	virtual std::string _printIP(libfwbuilder::IPService *srv);
	virtual std::string _printSrcPorts(libfwbuilder::Service *srv);
	virtual std::string _printDstPorts(libfwbuilder::Service *srv);

	virtual void _printMultiport(libfwbuilder::NATRule *r);

	virtual void _printAddr(libfwbuilder::Address  *o,bool print_mask=true,bool print_range=false);

	virtual void _printChainDirectionAndInterface(libfwbuilder::NATRule *r);

        void    splitRuleIfRuleElementIsDynamicInterface(libfwbuilder::NATRule *r,
                                                         std::vector<libfwbuilder::Rule*> *vr,
                                                         const std::string &re_type);

        virtual string _internalPrintNATRule(libfwbuilder::NATRule *rule);

	/**
	 * this inspector determines type of the NAT rule
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  NATRuleType );

	/**
	 * this inspector verifies correctness of the NAT rules
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  VerifyRules );

	/**
	 * this processor splits rule with multiple service objects in OSrv
	 * onto several rules
	 */
	DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt,  splitOnOSrv );

	/**
	 * this processor splits rule if one of the objects in tsrc is
         * interface with dynamic address
	 */
	DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, splitOnDynamicInterfaceInTSrc);

	/**
	 * this processor splits rule if one of the objects in odst is
         * interface with dynamic address
	 */
	DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, splitOnDynamicInterfaceInODst);

	/**
	 * this inspector fills translated service with the copy of original srv
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  fillTranslatedSrv );

	/**
	 * this inspector assigns NAT rules to interfaces 
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  AssignInterface );

	/**
	 * this inspector calls OSConfigurator to add virtual
	 * address to the firewall if it is needed for NAT rule
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  addVirtualAddress );

	/**
	 *  this inspector replaces references to the firewall in odst
	 *  with references to its external interfaces
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, ReplaceFirewallObjectsODst );

	/**
	 *  this inspector replaces references to the firewall in tsrc
	 *  with references to its external interfaces
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, ReplaceFirewallObjectsTSrc );

	/**
	 * this inspector distinguishes SNAT from Masquerading (can do
	 * this after firewall objects has been replaced with its
	 * interfaces and basic NAT rule type has been determined)
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, dynamicInterfaceInTSrc );

	/**
	 *  take care of dynamic interfaces in ODst  (if ODst contains
	 *  interface and its address is dynamic, replace it with any)
	 */
	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, dynamicInterfaceInODst );

        /**
         * this processor splits rule element if src or dst contains
         * address range. This inspector differs from the standard one
	 * in the base class NATCompiler
         */
        DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, addressRanges );

	/**
	 * split rules with more than one service object, so that each
	 * rule has services with the same protocol
	 */
	DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, splitServices );

	/**
	 * split rules so multiport module can be used
	 */
	DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, prepareForMultiport );

        /**
         * this processor deals with negation in OSrc 
         */
        DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, doOSrcNegation );

        /**
         * this processor deals with negation in ODst
         */
        DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, doODstNegation );

        /**
         * this processor deals with negation in OSrv 
         */
        DECLARE_NAT_RULE_PROCESSOR( NATCompiler_ipt, doOSrvNegation );


	/**
	 * decides what chain this rule should go to. 
	 */
        DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, decideOnChain );

	/**
	 * decides on "jump to" chain
	 */
        DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt, decideOnTarget );

	/**
	 *  this inspector prints single policy rule, assuming all
	 *  groups have been expanded, so source, destination and
	 *  service hold exactly one object each, and this object is
	 *  not a group.  Negation should also have been taken care of
	 *  before this method is called.
	 */
 	DECLARE_NAT_RULE_INSPECTOR( NATCompiler_ipt,  PrintAtomicRule );

	virtual std::string myPlatformName();

	std::map<const std::string,bool> chains;
	std::map<const std::string,bool> standard_chains;

	int chain_no;
	string getNewTmpChainName(libfwbuilder::NATRule *rule);
	string getNewChainName(libfwbuilder::NATRule *rule);

	public:

	NATCompiler_ipt(libfwbuilder::FWObjectDatabase *_db,
		       const std::string &fwname,
		       fwcompiler::OSConfigurator *_oscnf) : NATCompiler(_db,fwname,_oscnf) { chain_no=0; }


	virtual int  prolog();
	virtual void compile();
	virtual void epilog();
	
    };


}

#endif
