/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler_ipt.cc,v 1.26 2002/05/13 04:13:51 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "PolicyCompiler_ipt.hh"

#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/Policy.hh"

#include <iostream.h>
#include <streambuf.h>
#include <iomanip.h>
#include <fstream>
#include <sstream>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

string PolicyCompiler_ipt::myPlatformName() { return "iptables"; }

string PolicyCompiler_ipt::getNewTmpChainName(PolicyRule *rule)
{
    std::ostringstream str;

    str << "ptmp" << setw(3) << setfill('0') << chain_no;
    chain_no++;

    chains[str.str()]=true;
    return str.str();
}

string PolicyCompiler_ipt::getNewChainName(PolicyRule *rule)
{
    Interface *rule_iface = fw_interfaces[rule->getStr("interface_id")];

    std::ostringstream str;

    if (rule_iface)  str << rule_iface->getName() << "_";

    switch (rule->getDirection()) {
    case PolicyRule::Inbound:   str << "In_";  break;
    case PolicyRule::Outbound:  str << "Out_"; break;
    }
    str << "RULE_" << rule->getPosition();

    chain_no++;

    chains[str.str()]=true;

    return str.str();
}

static int cache_objects(FWObject *o, map<const string,FWObject*> &cache)
{
    int n=0;
    for (FWObject::iterator i=o->begin(); i!=o->end(); ++i) {
	if ( !(*i)->getId().empty() )
	    cache[(*i)->getId()]=(*i);
	n=n+1+cache_objects((*i),cache);
    }
    return n;
}

int PolicyCompiler_ipt::prolog()
{
    standard_chains["INPUT"]  =true;
    standard_chains["OUTPUT"] =true;
    standard_chains["FORWARD"]=true;
    standard_chains["RETURN"] =true;
    standard_chains["LOG"]    =true;
    standard_chains["ACCEPT"] =true;
    standard_chains["DROP"]   =true;
    standard_chains["REJECT"] =true;

    if (fw->getStr("platform")!="iptables") 
	abort("Unsupported platform " + fw->getStr("platform") );

    return PolicyCompiler::prolog();
}



/**
 * splits rule if logging is required and both src and dst are any
 */
void PolicyCompiler_ipt::Logging1::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrc *src=rule->getSrc();    assert(src);
    RuleElementDst *dst=rule->getDst();    assert(dst);

    if (rule->getLogging() && src->isAny() && dst->isAny()) {
	PolicyRule *r;
/*
 * split original rule onto two, set target to LOG in the first; turn
 * off logging and do not set target in the second.
 */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setStr("ipt_target","LOG");
	r->setLogging(false);
	vr->push_back(r);

	rule->setLogging(false);
	vr->push_back(rule);
    } else
	vr->push_back(rule);
}



/**
 * splits rule if logging is required and either src or dst is
 * not any
 */
void PolicyCompiler_ipt::Logging2::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrc *nsrc;
    RuleElementDst *ndst;
    RuleElementSrv *nsrv;

    if (rule->getLogging()) {
	string new_chain=getCompiler()->getNewChainName(rule);

	PolicyRule *r;
        FWOptions  *ruleopt;
/*
 * add copy of original rule, but turn off logging and set target
 * chain to new_chain.
 */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setStr("ipt_target",new_chain);
	r->setLogging(false);
	vr->push_back(r);

/*
 * need to add two rules with the same rule_label, direction=both, no
 * interface, no src, no srv and no dst. One of these new rules should
 * have target = LOG and another should inherit action and therefore
 * target from original rule. Both new rules go into chain new_chain.
 * In both rules turn off stateful inspection.
 *
 * keep interface information in the first one to be able to process
 * %I in log prefix
 *
 */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
        ruleopt =r->getOptionsObject();
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("ipt_target","LOG");
	r->setDirection( PolicyRule::Both );
	r->setLogging(false);
        ruleopt->setBool("stateless",true);
	vr->push_back(r);

	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
        ruleopt =r->getOptionsObject();
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	r->setStr("interface_id","nil");
	r->setDirection( PolicyRule::Both );
	r->setLogging(false);
        ruleopt->setBool("stateless",true);
	vr->push_back(r);

    } else
	vr->push_back(rule);
}

void PolicyCompiler_ipt::SrcNegation::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrc *srcrel=rule->getSrc();

/*   ! A  B  C  ACTION  */

    if (srcrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	srcrel->setNeg(false);


/*     any  B C  TMP_CHAIN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   A  any any  RETURN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
//	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}

void PolicyCompiler_ipt::DstNegation::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementDst *dstrel=rule->getDst();

/*   A  ! B  C  ACTION  */

    if (dstrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	dstrel->setNeg(false);


/*     A any C  TMP_CHAIN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   any B any  RETURN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
//	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}

void PolicyCompiler_ipt::SrvNegation::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrv *srvrel=rule->getSrv();

/*   A  B  !C  ACTION  */

    if (srvrel->getNeg()) {
	PolicyRule     *r;
	RuleElementSrc *nsrc;
	RuleElementDst *ndst;
	RuleElementSrv *nsrv;

	string new_chain=getCompiler()->getNewTmpChainName(rule);
	srvrel->setNeg(false);


/*     A B any  TMP_CHAIN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setLogging(false);
	r->setStr("ipt_target",new_chain);
	vr->push_back(r);

/* TMP_CHAIN   any  any  C  RETURN  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	r->setAction( PolicyRule::Return );
	r->setLogging(false);
	r->setStr("ipt_chain",new_chain);
//	r->setStr("interface_id","nil");
	vr->push_back(r);

/* TMP_CHAIN   any  any any  ACTION  */
	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,true) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	nsrc=r->getSrc();  nsrc->clearChildren();  nsrc->setAnyElement();
	ndst=r->getDst();  ndst->clearChildren();  ndst->setAnyElement();
	nsrv=r->getSrv();  nsrv->clearChildren();  nsrv->setAnyElement();
	r->setStr("ipt_chain",new_chain);
	if ( ! rule->getStr("ipt_target").empty() )
	    r->setStr("ipt_target",rule->getStr("ipt_target"));
//	r->setStr("interface_id","nil");
	vr->push_back(r);

    } else
	vr->push_back(rule);
}









void PolicyCompiler_ipt::InterfaceAndDirection::operator()(PolicyRule *rule) const
{
    Interface *rule_iface = getCompiler()->fw_interfaces[rule->getStr("interface_id")];
    
    if (rule->getDirection()==PolicyRule::Undefined) 
	rule->setDirection( PolicyRule::Both );

    if (rule_iface==NULL && rule->getDirection()==PolicyRule::Both) {
	rule->setStr("interface_id","nil");
	return;
    }

    if (rule_iface==NULL && (
	    rule->getDirection()==PolicyRule::Inbound ||
	    rule->getDirection()==PolicyRule::Outbound )
    ) throw FWException("Direction set without interface in rule "+rule->getStr("rule_label"));

}

/**
 *  re_type can be either RuleElementSrc::TYPENAME or RuleElementDst::TYPENAME
 */
void PolicyCompiler_ipt::splitRuleIfRuleElementIsForChainInput(PolicyRule *rule,
                                                               vector<Rule*> *vr,
                                                               const std::string &re_type)
{
    RuleElement    *re=RuleElement::cast(rule->getFirstByType(re_type));
    int nre=re->size();

    vector<FWObject*> cl;

    for (list<FWObject*>::iterator i1=re->begin(); nre>1 && i1!=re->end(); ++i1) {

	FWObject *o   = *i1;
	FWObject *obj = NULL;
	if (FWReference::cast(o)!=NULL) obj=objcache[o->getStr("ref")];
        Address *a=Address::cast(obj);
        assert(a!=NULL);

        IPAddress obj_addr=a->getAddress();

        if (complexMatch(a,fw)) {

	    cl.push_back(o);   // can not remove right now because remove invalidates iterator
            nre--;

	    PolicyRule  *new_rule= PolicyRule::cast(dbcopy->create(PolicyRule::TYPENAME,true) );
	    temp_ruleset->add(new_rule);
	    new_rule->duplicate(rule);
            RuleElement *new_re=RuleElement::cast(new_rule->getFirstByType(re_type));
	    new_re->clearChildren();
	    new_re->setAnyElement();
	    new_re->addRef( a );
	    vr->push_back(new_rule);
        }
        
    }
    if (!cl.empty()) {
        for (vector<FWObject*>::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  
            re->remove( (*i1) );
    }

    vr->push_back(rule);
}

void PolicyCompiler_ipt::splitIfInputChainObjectSrc::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    getCompiler()->splitRuleIfRuleElementIsForChainInput(rule,vr,RuleElementSrc::TYPENAME);
}

void PolicyCompiler_ipt::splitIfInputChainObjectDst::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    getCompiler()->splitRuleIfRuleElementIsForChainInput(rule,vr,RuleElementDst::TYPENAME);
}


void PolicyCompiler_ipt::splitIfIfaceAndDirectionBoth::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    Interface *rule_iface = getCompiler()->fw_interfaces[rule->getStr("interface_id")];

    if (rule_iface!=NULL && rule->getDirection()==PolicyRule::Both) {
	PolicyRule *r;

	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Inbound );
	vr->push_back(r);

	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Outbound );
	vr->push_back(r);
    } else
	vr->push_back(rule);
}

void PolicyCompiler_ipt::splitIfSrcAny::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    if ( ! rule->getStr("ipt_chain").empty() ) {
	vr->push_back(rule);
	return;
    }

    RuleElementSrc *src=rule->getSrc();

    if (src->isAny() && rule->getDirection()!=PolicyRule::Inbound ) {
	PolicyRule *r= PolicyRule::cast(
	    getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setStr("ipt_chain","OUTPUT");
	r->setDirection( PolicyRule::Outbound );
	vr->push_back(r);
    }
    vr->push_back(rule);  // add old rule in any case
}

void PolicyCompiler_ipt::splitIfDstAny::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    if ( ! rule->getStr("ipt_chain").empty() ) {
	vr->push_back(rule);
	return;
    }

    RuleElementDst *dst=rule->getDst();

    if (dst->isAny() && rule->getDirection()!=PolicyRule::Outbound ) {
	PolicyRule *r= PolicyRule::cast(
	    getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setStr("ipt_chain","INPUT");
	r->setDirection( PolicyRule::Inbound );
	vr->push_back(r);
    }
    vr->push_back(rule); // add old rule in any case
}

/*
 *  predicates that run before guarantee that when we call this one,
 *  firewall object, if it is in src or dst, is single object there
 */
void PolicyCompiler_ipt::checkSrcAndDst1::operator()(PolicyRule *rule) const
{
    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if (src->getId()!=getCompiler()->fw_id &&
	dst->getId()==getCompiler()->fw_id &&
	rule->getDirection()==PolicyRule::Outbound )
	throw FWException("direction can not be outbound when destination is firewall, in rule "+rule->getStr("rule_label"));
}

void PolicyCompiler_ipt::checkSrcAndDst2::operator()(PolicyRule *rule) const
{
    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if (src->getId()==getCompiler()->fw_id &&
	dst->getId()!=getCompiler()->fw_id &&
	rule->getDirection()==PolicyRule::Inbound )
	throw FWException("direction can not be inbound when source is firewall, in rule "+rule->getStr("rule_label"));
}

void PolicyCompiler_ipt::specialCaseWithFW1::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  
    if(src==NULL) throw("Broken SRC in rule "+rule->getStr("rule_label"));
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);
    if(dst==NULL) throw("Broken DST in rule "+rule->getStr("rule_label"));

    if (src->getId()==getCompiler()->fw_id &&
	dst->getId()==getCompiler()->fw_id &&
	rule->getDirection()== PolicyRule::Both ) {

	PolicyRule *r;

	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Inbound );
	vr->push_back(r);

	r= PolicyRule::cast(getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	getCompiler()->temp_ruleset->add(r);
	r->duplicate(rule);
	r->setDirection( PolicyRule::Outbound );
	vr->push_back(r);
    } else
	vr->push_back(rule);
}

void PolicyCompiler_ipt::decideOnChain::operator()(PolicyRule *rule) const
{
    if ( ! rule->getStr("ipt_chain").empty() ) return;

    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    rule->setStr("ipt_chain","FORWARD");

    if ( rule->getDirection()== PolicyRule::Inbound ) {
        if (getCompiler()->complexMatch(dst,getCompiler()->fw))
            rule->setStr("ipt_chain","INPUT");
        return;
    }

    if ( rule->getDirection()== PolicyRule::Outbound ) {
        if (getCompiler()->complexMatch(src,getCompiler()->fw))
            rule->setStr("ipt_chain","OUTPUT");
	return;
    }
/* direction == Both */

    if (!dst->isAny() && getCompiler()->complexMatch(dst,getCompiler()->fw)) {
        rule->setStr("ipt_chain","INPUT");
        return;
    }
    if (!src->isAny() && getCompiler()->complexMatch(src,getCompiler()->fw)) {
        rule->setStr("ipt_chain","OUTPUT");
        return;
    }
}

void PolicyCompiler_ipt::decideOnTarget::operator()(PolicyRule *rule) const
{
    if ( ! rule->getStr("ipt_target").empty() ) return; // already defined

    switch (rule->getAction()) {
    case PolicyRule::Accept:  rule->setStr("ipt_target","ACCEPT"); break;
    case PolicyRule::Deny:    rule->setStr("ipt_target","DROP");   break;
    case PolicyRule::Reject:  rule->setStr("ipt_target","REJECT"); break;
    case PolicyRule::Return:  rule->setStr("ipt_target","RETURN"); break;
    }
}

/*
 *  TODO:  may be make this optional, with checkbox in Firewall tab
 */
void PolicyCompiler_ipt::removeFW::operator()(PolicyRule *rule) const
{
    RuleElementSrc *srcrel=rule->getSrc();
    Address        *src   =srcrel->getFirst(true);  assert(src);
    RuleElementDst *dstrel=rule->getDst();
    Address        *dst   =dstrel->getFirst(true);  assert(dst);

    if ( rule->getStr("ipt_chain")=="INPUT" && 
	 dst->getId()==getCompiler()->fw_id ) {
	dstrel->clearChildren();
	dstrel->setAnyElement();
    }

    if ( rule->getStr("ipt_chain")=="OUTPUT" && 
	 src->getId()==getCompiler()->fw_id ) {
	srcrel->clearChildren();
	srcrel->setAnyElement();
    }
}

void PolicyCompiler_ipt::separatePortRanges::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrv *rel= rule->getSrv();

    if (rel->size()==1) {
	vr->push_back(rule);
	return;
    }

    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
	Service *s=Service::cast(o);
	assert(s!=NULL);

	if ( ! TCPService::isA(s) && ! UDPService::isA(s) ) continue;

	int srs, sre, drs, dre;
	    
	if (TCPService::isA(s)) {
	    TCPService *srv=TCPService::cast(s);
	    srs=srv->getInt("src_range_start");
	    sre=srv->getInt("src_range_end");
	    drs=srv->getInt("dst_range_start");
	    dre=srv->getInt("dst_range_end");
	}
	if (UDPService::isA(s)) {
	    UDPService *srv=UDPService::cast(s);
	    srs=srv->getInt("src_range_start");
	    sre=srv->getInt("src_range_end");
	    drs=srv->getInt("dst_range_start");
	    dre=srv->getInt("dst_range_end");
	}

	getCompiler()->normalizePortRange(srs,sre);
	getCompiler()->normalizePortRange(drs,dre);

	if (srs!=sre || drs!=dre) {
	    PolicyRule *r= PolicyRule::cast(
		getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	    getCompiler()->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementSrv *nsrv=r->getSrv();
	    nsrv->clearChildren();
	    nsrv->addRef( s );
	    vr->push_back(r);
	    services.push_back(s);
	} 

    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++) 
	rel->removeRef( (*i) );

    if (!rel->isAny())
	vr->push_back(rule);
}

/*
 * processor splitServices should have been called eariler, so now all
 * services in Srv are of the same type
 */
void PolicyCompiler_ipt::prepareForMultiport::operator()(PolicyRule *rule, vector<Rule*> *vr) const
{
    RuleElementSrv *rel= rule->getSrv();
    Service        *srv= rel->getFirst(true);

    if (rel->size()==1) {
	vr->push_back(rule);
	return;
    }

    if (IPService::isA(srv) || ICMPService::isA(srv)) {
/* multiport does  not support ip and icmp services, split the rule */
	for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
	    
	    Service *s=Service::cast( o );
	    assert(s);

	    PolicyRule *r= PolicyRule::cast(
		getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
	    getCompiler()->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementSrv *nsrv=r->getSrv();
	    nsrv->clearChildren();
	    nsrv->addRef( s );

	    vr->push_back(r);
	}
	return;
    }

    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	rule->setBool("ipt_multiport",true);

/* make sure we have no more than 15 ports */
	if (rel->size()>15) {

	    int n=0;
	    PolicyRule *r;
	    RuleElementSrv *nsrv;
	    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {

		FWObject *o= *i;
		if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
	    
		Service *s=Service::cast( o );
		assert(s);

		if (n==0) {
		    r= PolicyRule::cast(
			getCompiler()->dbcopy->create(PolicyRule::TYPENAME,false) );
		    getCompiler()->temp_ruleset->add(r);
		    r->duplicate(rule);
		    nsrv=r->getSrv();
		    nsrv->clearChildren();
		    vr->push_back(r);
		}
		nsrv->addRef( s );
		if (++n>=15) n=0;
	    }

	} else {
	    vr->push_back(rule);
	}
    }

}

void PolicyCompiler_ipt::prepareForMACMatching::operator()(PolicyRule *rule) const
{
    RuleElementSrc *rel= rule->getSrc();

    list<FWObject*> cl;
    for (FWObject::iterator i1=rel->begin(); i1!=rel->end(); ++i1) {
        FWObject *o= *i1;
        if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
        assert(o!=NULL);

	FWOptions *hopt;
	if (Host::isA(o) && (hopt=Host::cast(o)->getOptionsObject())!=NULL 
	    && hopt->getBool("use_mac_addr_filter") ) {
	    
	    cl.push_back(o);
	}
    }
    if ( ! cl.empty()) {
        for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1)  {
	    Host *h=Host::cast(*i1);
	    FWObjectTypedChildIterator j=h->findByType(Interface::TYPENAME);
	    if (j==j.end())
		throw FWException("MAC address match ordered for host object "+
				  h->getName()+", but it has no interfaces. Rule "+
				  rule->getStr("rule_label"));
	    rel->removeRef(h);
	    bool got_at_least_one_mac=false;
	    for ( ; j!=j.end(); ++j ) {
		Interface *iface=Interface::cast(*j);
		if ( ! iface->getPhysicalAddress().empty() ) {
		    got_at_least_one_mac=true;
		    iface->setBool("use_mac_addr",true);
		    rel->addRef(iface);
		}
	    }
	    if (! got_at_least_one_mac )
		throw FWException("MAC address match ordered for host object "+
				  h->getName()+", but none of its interfaces has MAC address defined. Rule "+
				  rule->getStr("rule_label"));
	}
    }
}

/**
 *-----------------------------------------------------------------------
 */
void PolicyCompiler_ipt::compile()
{
    cout << " Compiling policy for " << fw->getName() << " ..." <<  endl << flush;

    try {

	if (verbose) cout << "standard predefined rules ...\n" << flush;

        output << endl;

	if ( ! fwopt->getBool("accept_new_tcp_with_no_syn") ) {
	    output << "iptables -A OUTPUT  -p tcp ! --syn -m state --state NEW -j LOG\n";
	    output << "iptables -A OUTPUT  -p tcp ! --syn -m state --state NEW -j DROP\n";

	    output << "iptables -A INPUT   -p tcp ! --syn -m state --state NEW -j LOG\n";
	    output << "iptables -A INPUT   -p tcp ! --syn -m state --state NEW -j DROP\n";
                                          
	    output << "iptables -A FORWARD -p tcp ! --syn -m state --state NEW -j LOG\n";
	    output << "iptables -A FORWARD -p tcp ! --syn -m state --state NEW -j DROP\n";
	    output << endl;
	}

	if ( fwopt->getBool("log_all_dropped") ) {
	    output << "iptables -t drop -A DROPPING -j LOG " 
		   << _printGlobalLogParameters() << endl;
	    output << endl;
	}

	if ( fwopt->getBool("clamp_mss_to_mtu") ) {
	    output << "iptables -A FORWARD -p tcp --tcp-flags SYN,RST SYN -j TCPMSS --clamp-mss-to-pmtu" 
		   << endl;
	    output << endl;
	}

	if ( fwopt->getBool("accept_established") ) {
	    output << "iptables -A INPUT   -m state --state ESTABLISHED,RELATED -j ACCEPT"
		   << endl;
	    output << "iptables -A OUTPUT  -m state --state ESTABLISHED,RELATED -j ACCEPT"
		   << endl;
	    output << "iptables -A FORWARD -m state --state ESTABLISHED,RELATED -j ACCEPT"
		   << endl;
	    output << endl;
	}

	Compiler::compile();

        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "fill in interface and direction ..\n" << flush;
	for_each_rule( InterfaceAndDirection(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "process negation in Src ...\n" << flush;
	for_each_rule( SrcNegation(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "process negation in Dst ...\n" << flush;
	for_each_rule( DstNegation(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "process negation in Srv ...\n" << flush;
	for_each_rule( SrvNegation(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

//	if (verbose) cout << "split rule if interface is defined and direction is Both ...\n" << flush;
//	for_each_rule( splitIfIfaceAndDirectionBoth(this) );
//        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "process logging...\n" << flush;
//	for_each_rule( Logging1(this) );
	for_each_rule( Logging2(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

/* SrcAndDst */

	if ( fwopt->getBool("firewall_is_part_of_any_and_networks") ) {
            if (verbose) cout << "split rule if Src is any ...\n" << flush;
            for_each_rule( splitIfSrcAny(this) );
            if (debug_rule>=0) for_each_rule( debugRule(this) );

            if (verbose) cout << "split rule if Dst is any ...\n" << flush;
            for_each_rule( splitIfDstAny(this) );
            if (debug_rule>=0) for_each_rule( debugRule(this) );
        }

        if (verbose) cout << "expand groups ...\n" << flush;
        for_each_rule( ExpandGroups(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rule if firewall or broadcast is in Src ..\n" << flush;
	for_each_rule( splitIfInputChainObjectSrc(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rule if firewall or broadcast is in Dst ..\n" << flush;
	for_each_rule( splitIfInputChainObjectDst(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

/*  at this point in all rules where firewall is in either src or dst,
 *  firewall is single object in that rule element. Other rule
 *  elements may contain multiple objects yet
 */

	if (verbose) cout << "process case when firewall is in src and dst ...\n" << flush;
	for_each_rule( specialCaseWithFW1(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );


//	if (verbose) cout << "check for illegal combination of src and dst ..\n" << flush;
//	for_each_rule( checkSrcAndDst1(this) );
//	for_each_rule( checkSrcAndDst2(this) );
//      if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "decide on chain ...\n" << flush;
	for_each_rule( decideOnChain(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "decide on target ...\n" << flush;
	for_each_rule( decideOnTarget(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "remove firewall obj. if chain is INPUT or OUTPUT ...\n" << flush;
	for_each_rule( removeFW(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "prepare for MAC address matches ...\n" << flush;
	for_each_rule( prepareForMACMatching(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand objects with multiple addresses ...\n" << flush;
	for_each_rule( ExpandMultipleAddresses(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand address range objects  ...\n" << flush;
	for_each_rule( addressRanges(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rules with different protocols ...\n" << flush;
	for_each_rule( splitServices(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "separate TCP services with flags ...\n" << flush;
	for_each_rule( separateTCPWithFlags(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "verify custom services for this platform ...\n" << flush;
	for_each_rule( verifyCustomServices(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "separate TCP/UDP services with port ranges ...\n" << flush;
	for_each_rule( separatePortRanges(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );



	if (verbose) cout << "optimizer pass 1 ...\n" << flush;
	for_each_rule( optimize1(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "optimizer pass 2 ...\n" << flush;
	for_each_rule( optimize1(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );




	if (verbose) cout << "prepare for using multiport ...\n" << flush;
	for_each_rule( prepareForMultiport(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "convert to atomic rules ...\n" << flush;
	for_each_rule( ConvertToAtomicForAddresses(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

//	if (verbose) cout << "detect rule shading ...\n" << flush;
//	for_each_rule( DetectShading(this) );
//      if (debug_rule>=0) for_each_rule( debugRule(this) );

    } catch (FWException &ex) {
	abort(ex.toString());
    }
}

string PolicyCompiler_ipt::_internalPrintPolicyRule(libfwbuilder::PolicyRule *rule)
{
    return PolicyCompiler::_internalPrintPolicyRule(rule)+
        " "+rule->getStr("ipt_chain")+" "+rule->getStr("ipt_target");
}

void PolicyCompiler_ipt::epilog()
{
    try {
	if (verbose) cout << "generate iptables code ... \n" << flush;

	output << "#  temporary chains\n#\n";

	map<string,bool>::iterator i;
	for (i=chains.begin(); i!=chains.end(); ++i)
	    output << "iptables -N " << (*i).first << endl;

	output << "#\n#\n";

	string current_rule_label="";

	for (list<FWObject*>::iterator k=combined_ruleset->begin(); k!=combined_ruleset->end(); ++k) {
	    PolicyRule *r = PolicyRule::cast( *k );
	    string rl=r->getStr("rule_label");
	    if (rl!=current_rule_label) {
		output << "# " << endl;
		output << "# Rule " << rl << endl;
		output << "# " << endl;

		string    comm=r->getComment();
		string::size_type c1,c2;
		c1=0;
		while ( (c2=comm.find('\n',c1))!=string::npos ) {
		    output << "# " << comm.substr(c1,c2-c1) << endl;
		    c1=c2+1;
		}
		output << "# " << comm.substr(c1) << endl;
		output << "# " << endl;

		current_rule_label=rl;
	    }
	    PrintRule(this)(r);
	}

	output << "#" << endl;
	output << "#" << endl;
/*
	output << "#  Fallback rules" << endl;
	output << "#" << endl;
	output << "iptables -A INPUT    -j DROP"  << endl;		       
	output << "iptables -A OUTPUT   -j DROP"  << endl;		       
	output << "iptables -A FORWARD  -j DROP"  << endl;
	output << endl;
*/

/* Turn on packet forwarding if we have to */

    string s=fwopt->getStr("linux24_ip_forward");
    if (!s.empty())
	output << "echo " << s << " > /proc/sys/net/ipv4/ip_forward\n\n";
    else
	output << "echo \"$FWD\" > /proc/sys/net/ipv4/ip_forward\n\n";

    cout << " Policy compiled successfully \n" << flush;

    } catch (FWException ex) {
	abort(ex.toString());
    }
}
