/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler_ipt.hh,v 1.17 2002/05/10 03:09:34 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _POLICYCOMPILER_IPT_HH
#define _POLICYCOMPILER_IPT_HH

#include "fwcompiler/PolicyCompiler.hh"


namespace libfwbuilder {
class IPService;
class ICMPService;
class TCPService;
class UDPService;
};

namespace fwcompiler {


    class PolicyCompiler_ipt : public PolicyCompiler {

	protected:


	virtual void _printSrcService(libfwbuilder::RuleElementSrv  *o);
	virtual void _printDstService(libfwbuilder::RuleElementSrv  *o);
	virtual void _printProtocol(libfwbuilder::Service *srv);

	virtual std::string _printPorts(int rs,int re);
	virtual std::string _printSrcPorts(libfwbuilder::Service *srv);
	virtual std::string _printDstPorts(libfwbuilder::Service *srv);
	virtual std::string _printICMP(libfwbuilder::ICMPService *srv);
	virtual std::string _printIP(libfwbuilder::IPService *srv);

	virtual std::string _printTCPFlags(libfwbuilder::TCPService *srv);

	virtual void _printSrcAddr(libfwbuilder::RuleElementSrc  *o);
	virtual void _printDstAddr(libfwbuilder::RuleElementDst  *o);
	virtual void _printAddr(libfwbuilder::Address  *o);

	virtual void _printChain(libfwbuilder::PolicyRule *r);
	virtual void _printTarget(libfwbuilder::PolicyRule *r);
	virtual void _printDirectionAndInterface(libfwbuilder::PolicyRule *r);
	virtual void _printMultiport(libfwbuilder::PolicyRule *r);

	virtual std::string _printGlobalLogParameters();
	virtual std::string _printLogParameters(libfwbuilder::PolicyRule *r);
	virtual std::string _printActionOnReject(libfwbuilder::PolicyRule *r);

	virtual void _checkAndCreateNewChain(libfwbuilder::PolicyRule *r);

        void    splitRuleIfRuleElementIsForChainInput(libfwbuilder::PolicyRule *r,
                                                      std::vector<libfwbuilder::Rule*> *vr,
                                                      const std::string &re_type);

        void _optimizeForRuleElement(libfwbuilder::PolicyRule         *rule,
                                     std::vector<libfwbuilder::Rule*> *vr,
                                     const std::string  &re_type);

        /**
         * prints rule in some universal format (close to that visible
         * to user in the GUI). Used for debugging purposes. This method
         * calls PolicyCompiler::_internalPrintPolicyRule and then adds
         * chain and target at the end of the printed line
         */
        virtual string _internalPrintPolicyRule(libfwbuilder::PolicyRule *rule);


	/**
	 * splits rule if logging is required and both src and dst are any
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, Logging1 );

	/**
	 * splits rule if logging is required and either src or dst is
	 * not any
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, Logging2 );

	/**
	 * processes rules with negation in Src
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, SrcNegation );

	/**
	 * processes rules with negation in Dst
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, DstNegation );

	/**
	 * processes rules with negation in Srv
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, SrvNegation );

	/**
	 *  this rule processor splits rule if one of the objects in
	 *  Src is firewall itself or a broadcast address for one of
	 *  its interfaces or multicast (multicast support is not
	 *  implemented yet). This is needed to properly choose
	 *  direction and chain later. Src may have multiple objects
	 *  when this processor is called.  For firewall the iptables
	 *  code compiler generates should go into OUTPUT chain, while
	 *  for all other objects it should go into FORWARD
	 *  chain. That is why we need to split the rule and place
	 *  firewall object in the src of the rule of its own.
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, splitIfInputChainObjectSrc );


	/**
	 *  this rule processor splits rule if one of the objects in Dst
	 *  is firewall itself. This is needed to properly choose direction 
	 *  and chain later. See comment in splitIfInputChainObjectSrc.
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, splitIfInputChainObjectDst );


        /**
	 *  This inspector verifies combination of interface and
	 *  direction and fills interface and direction. After this
	 *  predicate it is guaranteed that both interface and
	 *  direction have some value. In certain situations interface
	 *  ID may be set to "nil" though (e.g. global policy rules).
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, InterfaceAndDirection );

	/**
	 * splits rule onto two if interface is defined and direction is Both
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, splitIfIfaceAndDirectionBoth );

	/**
	 * split rule if Src==any 
	 * 
	 * This is special case since we assume that "any" includes
	 * also a firewall object. Packets headed to or from the
	 * firewall must be inspected by INPUT or OUTPUT chain, while
	 * packets crossing the firewall are inspected by FORWARD
	 * chain. If we assume that "any" also includes firewall
	 * itself, then we need to generate code for both FORWARD and
	 * INPUT/OUTPUT chains from the same rule. This processor
	 * splits the rule onto two and sets chain and direction in
	 * the second copy appropriately. It preserve original src and
	 * dst in both copies, it only changes chain and direction in
	 * the second copy.
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, splitIfSrcAny );

	/**
	 * split rule if Dst==any. See comment in splitIfSrcAny
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, splitIfDstAny );

	/**
	 * checks for illegal combination of src, dst and direction
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, checkSrcAndDst1 );

	/**
	 * checks for illegal combination of src, dst and direction
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, checkSrcAndDst2 );

	/**
	 * splits rule if firewall is in src and dst
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, specialCaseWithFW1 );

	/**
	 * decides what chain this rule should go to. At this point
	 * src and dst should be already processed by previous
	 * predicates in a such way that if firewall object is in src
	 * or dst, then it should be a single object there.
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, decideOnChain );

	/**
	 * decides on "jump to" chain
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, decideOnTarget );

	/**
	 * If chain has been determined to be INPUT or OUTPUT, we can
	 * remove firewall object from dst or src (resp.) NB: we can
	 * remove only reference to the whole firewall. We DO NOT
	 * remove reference to its interface or (in the future)
	 * address objects under interfaces
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, removeFW );

	/**
	 * separate TCP/UDP services with port ranges (can not be used with multiport)
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, separatePortRanges );

	/**
	 * optimize rules in case we deal with one or few objects in
	 * one rule element and lots of objects in the other two
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, optimize1 );

	/**
	 * split rules so multiport module can be used
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_ipt, prepareForMultiport );

	/**
	 * prepares for MAC matching (if host object has multiple interfaces
	 * with MAC addresses defined, then this processor replaces reference
	 * to the host with references to its interfaces )
	 *
	 * This inspector should be to called before
	 * ExpandMultipleAddresses because the latter replaces
	 * references to Host objects with references to their
	 * interfaces (if any) but prepareForMACMatching needs to
	 * access host's options
	 */
	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, prepareForMACMatching );

	/**
	 *  this inspector prints single policy rule, assuming all
	 *  groups have been expanded, so source, destination and
	 *  service hold exactly one object each, and this object is
	 *  not a group.  Negation should also have been taken care of
	 *  before this method is called.
	 */
 	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_ipt, PrintRule );

	virtual std::string myPlatformName();

	std::map<const std::string,bool> chains;
	std::map<const std::string,bool> standard_chains;

	int chain_no;
	string getNewTmpChainName(libfwbuilder::PolicyRule *rule);
	string getNewChainName(libfwbuilder::PolicyRule *rule);

	public:

	PolicyCompiler_ipt(libfwbuilder::FWObjectDatabase *_db,
			  const std::string &fwname,
			  fwcompiler::OSConfigurator *_oscnf) : PolicyCompiler(_db,fwname,_oscnf) { chain_no=0; }


	virtual int  prolog();
	virtual void compile();
	virtual void epilog();
	
    };


}

#endif
