/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: ipt.cc,v 1.6 2002/05/10 03:09:34 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

#include <fstream>
#include <iostream>
#include <algorithm>
#include <functional>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "PolicyCompiler_ipt.hh"
#include "NATCompiler_ipt.hh"

#include "OSConfigurator_linux24.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/XMLTools.hh"
#include "fwbuilder/FWException.hh"

#include "fwcompiler/manifest.hh"

#ifdef HAVE_GETOPT_H
  #include <getopt.h>
#else
  #ifdef __MINGW32__
    #include <getopt.h>
  #else
    #include <stdlib.h>
  #endif
#endif


static const char      *filename       = NULL;
static const char      *wdir           = NULL;
static const char      *mfile          = NULL;
static const char      *fwobjectname   = NULL;
static int              dl             = 0;
static int              drp            = -1;
static int              drn            = -1;
static int              verbose        = 0;

using namespace libfwbuilder;
using namespace fwcompiler;

class UpgradePredicate: public XMLTools::UpgradePredicate
{
    public:
    virtual bool operator()(const string &msg) const 
    { 
	cout << "Data file has been created in the old version of Firewall Builder. Use fwbuilder GUI to convert it." << endl;
	return false;
    }
};
    
void usage(const char *name)
{
    cout << "Firewall Builder:  policy compiler for Linux 2.4.xx iptables" << endl;
    cout << "Version " << VERSION << endl;
    cout << "Usage: " << name << " [-x level] [-v] [-V] [-f filename.xml] [-d destdir] [-m manifest] firewall_object_name" << endl;
}

/**
 *  TODO:  move this to fwcompiler
 */
Firewall* findFirewallByName(FWObjectDatabase* db,const string &name)
{
    FWObject *fwgrp=db->getById( FWObjectDatabase::db->std.FirewallsId , true );
    assert(fwgrp!=NULL);

    FWObject::const_iterator i=
	std::find_if( fwgrp->begin(), fwgrp->end(), FWObjectNameEQPredicate(name) );

    return (i==fwgrp->end())?NULL:Firewall::cast( *i );
}

int main(int argc, char * const *argv)
{   

    if (argc<=1)
    {
        usage(argv[0]);
        exit(1);
    }

    int   opt;

    while( (opt=getopt(argc,argv,"x:vVf:d:m:")) != EOF )
    {
        switch(opt)
        {
        case 'd':
            wdir = strdup(optarg);
            break;
        case 'm':
            mfile = strdup(optarg);
            break;
        case 'f':
            filename = strdup(optarg);
            break;
        case 'x':
            if (*optarg=='r') {
                ++optarg;
                drp  = atoi(optarg);
            } else
                if (*optarg=='n') {
                    ++optarg;
                    drn  = atoi(optarg);
                } else
                    dl++;  // increase debug level
            break;
        case 'v':
            verbose++;
            break;
        case 'V':
            usage(argv[0]);
            exit(1);
        }
    }
    
    if((argc-1) != optind)
    {
        usage(argv[0]);
        exit(1);
    }

    fwobjectname = strdup( argv[optind++] );


    if (wdir==0) 	wdir="./";

    if(chdir(wdir))   {
	cerr << "Can't change to: " << wdir << endl;
	exit(1);
    }


    try   {

	/* init threads */
	g_thread_init(NULL);
    
        libfwbuilder::init();

	/* Initialize XML parser  */
	XMLTools::initXMLTools();

	/* create database */
	new FWObjectDatabase();

	/* load the data file */
	UpgradePredicate upgrade_predicate; 

	if (verbose) cout << " *** Loading data ...";
	FWObjectDatabase::db->load(filename,  &upgrade_predicate);
	if (verbose) cout << " done\n";


	string   fw_file_name=string(fwobjectname)+".fw";
	ofstream fw_file(fw_file_name.c_str());

	/* Review firewall and OS options and generate commands */
	Firewall*  fw=findFirewallByName(FWObjectDatabase::db, fwobjectname);
	FWOptions* options=fw->getOptionsObject();
	string s;

	string firewall_dir=options->getStr("firewall_dir");

	bool debug=options->getBool("debug");
	string shell_dbg=(debug)?"-x":"" ;
	string pfctl_dbg=(debug)?"-v":"";

	fw_file << "#!/bin/sh " << shell_dbg << endl << endl;

	char           timestr[256];
	time_t         tm;

	tm=time(NULL);
	strcpy(timestr,ctime(&tm));
	timestr[ strlen(timestr)-1 ]='\0';
    
        char* user_name=(char*)g_get_user_name();
        if (user_name==NULL) {
	    cerr << "Can't figure out your user name, aborting" << endl;
	    exit(1);
	}

	fw_file << "#\n\
#  This is automatically generated file. DO NOT MODIFY !\n\
#\n\
#  Firewall Builder  fwb_ipt v" << VERSION << " \n\
#\n\
#  Generated " << timestr << " " << tzname[0] << " by " 
		<< user_name << "\n#\n#\n#\n#\n";

	if ( firewall_dir.empty() ) firewall_dir="/etc";

	fw_file << "cd " << firewall_dir << " || exit 1" << endl << endl;
	fw_file << "if [ -x /usr/bin/logger ]; then" << endl;
	fw_file << "  logger ";

	string log_level=options->getStr("log_level");
	if ( ! log_level.empty() ) 
	    fw_file << "-p " << log_level << " ";
	
	fw_file << "\"Activating firewall script " 
		<< fw_file_name << " generated " 
		<< timestr << " " << tzname[0] << " by " 
		<< user_name << "\"" << endl;

	fw_file << "fi" << endl;

	string env_path=options->getStr("script_env_path");
	if ( ! env_path.empty() )
	    fw_file << "PATH=" << env_path << endl;

	fw_file << endl;

/*
 * Process firewall options, build OS network configuration script
 */
	OSConfigurator *oscnf=NULL;
	if (fw->getStr("host_OS")=="linux24")   oscnf=new OSConfigurator_linux24(FWObjectDatabase::db , fwobjectname);
	if (oscnf==NULL)
	    throw FWException("Unrecognized host OS "+fw->getStr("host_OS"));
	oscnf->processFirewallOptions();




	fw_file << "iptables -P OUTPUT  DROP" << endl;
	fw_file << "iptables -P INPUT   DROP" << endl;
	fw_file << "iptables -P FORWARD DROP" << endl;
	fw_file << endl;

/*
 *    flush all the rules which existed before
 */

	fw_file << "\n\
\n\
cat /proc/net/ip_tables_names | while read table; do\n\
  iptables -t $table -L -n | while read c chain rest; do\n\
      if test \"X$c\" = \"XChain\" ; then\n\
        iptables -t $table -F $chain\n\
      fi\n\
  done\n\
  iptables -t $table -X\n\
done\n\
\n";
    
	fw_file << endl;;


/*
 * create compilers and run the whole thing 
 */
        Manifest manifest;
	Action a;

	a.action     = "COPY";
	a.filename   = fw_file_name;
	a.parameters = firewall_dir;
	manifest.push_back(a);


	PolicyCompiler_ipt c( FWObjectDatabase::db , fwobjectname , oscnf );

	c.setDebugLevel( dl );
	c.setDebugRule(  drp );
	c.setVerbose( verbose );

	if ( c.prolog() > 0 ) {
	    c.compile();
	    c.epilog();
	}


	NATCompiler_ipt n( FWObjectDatabase::db , fwobjectname , oscnf );

	n.setDebugLevel( dl );
	n.setDebugRule(  drn );
	n.setVerbose( verbose );

	if ( n.prolog() > 0 ) {
	    n.compile();
	    n.epilog();
	}
/*
 * now write generated scripts to files
 */
	fw_file << oscnf->getCompiledScript();
	fw_file << n.getCompiledScript();
	fw_file << c.getCompiledScript();
	fw_file << endl;

/* TODO : need to restore state of ip_forward _AFTER_ all iptables commands */
	fw_file.close();


	if(mfile)
            manifest.save(string(mfile), S_IRUSR|S_IWUSR);
        
        return 0;

    } catch(FWException &ex)  {
	cerr << ex.toString() << endl;
        return 1;
    } catch (std::string s) {
	cerr << s;
        return 1;
    } catch (std::exception ex) {
	cerr << ex.what();
        return 1;
    } catch (...) {
	cerr << "Unsupported exception";
        return 1;
    }

}









