/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_pf.cc,v 1.28 2002/05/13 04:13:51 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "NATCompiler_pf.hh"

#include "fwcompiler/OSConfigurator.hh"

#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/NAT.hh"
#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/Service.hh"
#include "fwbuilder/Host.hh"
#include "fwbuilder/Network.hh"

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

string NATCompiler_pf::myPlatformName() { return "pf"; }


int NATCompiler_pf::prolog()
{

    int n=NATCompiler::prolog();

    if ( n>0 ) {

	bool found_ext=false;
	list<FWObject*> l2=fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    if ( iface->isExt() ) found_ext=true;
	}

	if (!found_ext)
	    throw FWException("At least one interface should be marked as external, can not configure NAT");
    }

/* pseudo-host with ip address 127.0.0.1  We'll use it for redirection NAT rules */
    FWObject    *grp;
    loopback_address=Host::cast(dbcopy->create(Host::TYPENAME,false) );
    loopback_address->setStr("address","127.0.0.1");
    grp=dbcopy->getById( dbcopy->std.HostsId , true );
    assert(grp!=NULL);
    grp->add(loopback_address);
    objcache[loopback_address->getId()]=loopback_address; // to keep cache consistent

    return n;
}

string NATCompiler_pf::_internalPrintNATRule(libfwbuilder::NATRule *rule)
{
    return NATCompiler::_internalPrintNATRule(rule)+
        " "+rule->getStr("interface_id")+
        " (type="+rule->getRuleTypeAsString()+")";
}


void NATCompiler_pf::NATRuleType::operator()(NATRule *rule) const
{
    Address  *osrc=rule->getOSrc()->getFirst(true);
    Address  *odst=rule->getODst()->getFirst(true);

    Address  *tsrc=rule->getTSrc()->getFirst(true);
    Address  *tdst=rule->getTDst()->getFirst(true);

    if (   tsrc->isAny() &&   tdst->isAny() ) {
	rule->setRuleType(NATRule::NONAT);
	return;
    }

    if ( ! tsrc->isAny() && tdst->isAny() ) {
        rule->setRuleType(NATRule::SNAT);
	return;
    }

    if (   tsrc->isAny() && ! tdst->isAny() ) {
        if ( tdst==getCompiler()->fw ) rule->setRuleType(NATRule::Redirect);
        else                           rule->setRuleType(NATRule::DNAT);
        return;
    }

    throw FWException("Unsupported translation. Rule: "+rule->getStr("rule_label"));
}

void NATCompiler_pf::VerifyRules::operator()(NATRule *rule) const
{
    RuleElementOSrc  *osrc=rule->getOSrc();  assert(osrc);
    RuleElementODst  *odst=rule->getODst();  assert(odst);
    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);

    RuleElementTSrc  *tsrc=rule->getTSrc();  assert(tsrc);
    RuleElementTDst  *tdst=rule->getTDst();  assert(tdst);
    RuleElementTSrv  *tsrv=rule->getTSrv();  assert(tsrv);

    if (rule->getRuleType()==NATRule::DNAT && odst->size()!=1)
	throw FWException("There should be no more than one object in original destination in the rule "+rule->getStr("rule_label"));

    if (rule->getRuleType()==NATRule::SNAT && tsrc->size()!=1)
	throw FWException("There should be no more than one object in translated source in the rule "+rule->getStr("rule_label"));

    if (osrv->getNeg())
        throw FWException("Negation in original service is not supported. Rule "+rule->getStr("rule_label"));

    if (rule->getRuleType()==NATRule::DNAT && osrv->isAny()) 
	throw FWException("Service must be specified for destination translation rule. Rule "+rule->getStr("rule_label"));

    if (tsrc->getNeg())
        throw FWException("Can not use negation in translated source. Rule "+rule->getStr("rule_label"));

    if (tdst->getNeg())
        throw FWException("Can not use negation in translated destination. Rule "+rule->getStr("rule_label"));

    if (tsrv->getNeg())
        throw FWException("Can not use negation in translated service. Rule "+rule->getStr("rule_label"));

    if (tsrv->size()!=1) 
	throw FWException("Translated service should be 'Original' or should contain single object. Rule: "+rule->getStr("rule_label"));

    FWObject *o=tsrv->front();
    if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];

    if ( Group::cast(o)!=NULL)
	throw FWException("Can not use group in translated service. Rule "+rule->getStr("rule_label"));


    if (rule->getRuleType()==NATRule::SNAT ) {
        if ( tsrc->size()!=1)
            throw FWException("There should be no more than one object in translated source in the rule "+rule->getStr("rule_label"));

        Address* o1=tsrc->getFirst(true);
        if ( Network::cast(o1)!=NULL || AddressRange::cast(o1)!=NULL )
            throw FWException("Can not use network or address range object in translated source. Rule "+rule->getStr("rule_label"));
    }

    if (rule->getRuleType()==NATRule::DNAT || rule->getRuleType()==NATRule::Redirect ) {
        if ( tdst->size()!=1)
            throw FWException("There should be no more than one object in translated destination in the rule "+rule->getStr("rule_label"));

        Address* o1=tdst->getFirst(true);
        if ( Network::cast(o1)!=NULL || AddressRange::cast(o1)!=NULL )
            throw FWException("Can not use network or address range object in translated destination. Rule "+rule->getStr("rule_label"));
    }


    if (rule->getRuleType()==NATRule::SNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(osrc->getFirst(true));
        Network *a2=Network::cast(tsrc->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException("Original and translated source should both be networks of the same size . Rule "+rule->getStr("rule_label"));
    }

    if (rule->getRuleType()==NATRule::DNetnat && !tsrc->isAny() ) {
        Network *a1=Network::cast(odst->getFirst(true));
        Network *a2=Network::cast(tdst->getFirst(true));
        if ( a1==NULL || a2==NULL ||
             a1->getNetmask().getLength()!=a2->getNetmask().getLength() )
            throw FWException("Original and translated destination should both be networks of the same size . Rule "+rule->getStr("rule_label"));
    }
}

void NATCompiler_pf::splitOnOSrv::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    RuleElementOSrv  *osrv=rule->getOSrv();  assert(osrv);
    if (osrv->size()!=1) {

	for(list<FWObject*>::iterator i=osrv->begin(); i!=osrv->end(); ++i) {
	    FWObject *o= *i;
//	    if (FWReference::cast(o)!=NULL) o=FWReference::cast(o)->getPointer();
	    if (FWReference::cast(o)!=NULL) o=getCompiler()->objcache[o->getStr("ref")];
	    Service *s=Service::cast( o );
	    assert(s);

	    NATRule *r= NATRule::cast(getCompiler()->
				      dbcopy->create(NATRule::TYPENAME,true) );
	    getCompiler()->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementOSrv *nosrv=r->getOSrv();
	    nosrv->clearChildren();

	    nosrv->addRef( s );

	    vr->push_back( r );
	}
    } else
	vr->push_back(rule);
}

void NATCompiler_pf::fillTranslatedSrv::operator()(NATRule *rule) const
{
    Service  *osrv_o=rule->getOSrv()->getFirst(true);
    Service  *tsrv_o=rule->getTSrv()->getFirst(true);

    if ( ! osrv_o->isAny()  && tsrv_o->isAny() ) {
	RuleElementTSrv  *tsrv=rule->getTSrv();
	tsrv->addRef(osrv_o);
    }
}

void NATCompiler_pf::addVirtualAddress::operator()(NATRule *rule) const
{
    Address *a=NULL;
    if (rule->getRuleType()==NATRule::SNAT) a=rule->getTSrc()->getFirst(true);
    else
	if (rule->getRuleType()==NATRule::DNAT) a=rule->getODst()->getFirst(true);
	else  return;
    assert(a!=NULL);

    if (a->getId()!=getCompiler()->fw_id ) {

	list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    if ( a->getAddress() == iface->getAddress() ) return;
	}
	getCompiler()->osconfigurator->addVirtualAddressForNAT( a->getAddress() );
    }
}

void NATCompiler_pf::AssignInterface::operator()(NATRule *rule, vector<Rule*> *vr) const
{
    Address *a=NULL;

    switch (rule->getRuleType() ) {

#ifdef ASSIGN_IFACE_FOR_NONAT
    case NATRule::Continue:
    case NATRule::NONAT:
    {
/* use heuristic to assign nonat rule to interfaces */
        Interface *iface;

        a=rule->getODst()->getFirst(true);
        iface=getCompiler()->findInterfaceFor(a,getCompiler()->fw);
        if (iface==NULL || iface->isLoopback()) { 
            a=rule->getOSrc()->getFirst(true);
            iface=getCompiler()->findInterfaceFor(a,getCompiler()->fw);
        }
        if (iface!=NULL && !iface->isLoopback()) { 
            rule->setStr("interface_id", iface->getId() );
            vr->push_back( rule );
            return;
        }
// slip into Redirect case to assign rule to all interfaces
    }
#endif
    case NATRule::Continue:
    case NATRule::NONAT:
    case NATRule::Redirect: 
    {
/* assign rule to all interfaces, except loopback */

        FWObjectTypedChildIterator j=getCompiler()->fw->findByType(Interface::TYPENAME);
        if (j==j.end()) {
            vr->push_back( rule );  // no interfaces ?!
            return;
        }
        for ( ; j!=j.end(); ++j ) {
            Interface *iface=Interface::cast(*j);
            assert(iface);
            if ( iface->isLoopback()) continue;

            NATRule *r= NATRule::cast(getCompiler()->dbcopy->create(NATRule::TYPENAME,true) );
            getCompiler()->temp_ruleset->add(r);
            r->duplicate(rule);
            r->setStr("interface_id", iface->getId() );
            vr->push_back( r );
        }
        return;
    }

    case NATRule::SNAT:
    case NATRule::SNetnat:
	a=rule->getTSrc()->getFirst(true);
        break;

    case NATRule::DNAT:
    case NATRule::DNetnat:
	a=rule->getODst()->getFirst(true);
        break;

    }

    if (a!=NULL) {

        Interface *iface=getCompiler()->findInterfaceFor(a,getCompiler()->fw);
        if (iface!=NULL && !iface->isLoopback()) { 
            rule->setStr("interface_id", iface->getId() );
            vr->push_back(rule);
            return;
        }
    }
    throw FWException("Could not assign NAT rule to the interface. Perhaps one of the objects has address which does not belong to any subnet the firewall has interface on. Rule: "+rule->getStr("rule_label"));
}

/*
 *  I assume that there is always only one object in ODst, TSrc and TDst
 *  rule elements. This should have been assured by inspector VerifyRules
 */
void NATCompiler_pf::ReplaceFirewallObjectsODst::operator()(NATRule *rule) const
{
    list<FWObject*> cl;
    RuleElementODst *rel;
    Address         *obj=NULL;

    rel=rule->getODst();      assert(rel);
    obj=rel->getFirst(true);  assert(obj);

    if (obj->getId()==getCompiler()->fw_id ) {

	list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	    Interface *interface_=Interface::cast(*i);
	    if (interface_->isExt()) 
		cl.push_back(interface_);
	}
	if ( ! cl.empty() ) {
	    rel->clearChildren();

	    for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
	    {
		rel->addRef( *i1 );
	    }
	}
    }
}

/*
 *  I assume that there is always only one object in ODst, TSrc and TDst
 *  rule elements. This should have been assured by inspector VerifyRules
 */
void NATCompiler_pf::ReplaceFirewallObjectsTSrc::operator()(NATRule *rule) const
{
    list<FWObject*> cl;
    RuleElementTSrc *rel;
    Address         *obj=NULL;

    switch (rule->getRuleType()) {

    case NATRule::Masq:   return;
    default:
	rel=rule->getTSrc();      assert(rel);
	obj=rel->getFirst(true);  assert(obj);

	if (obj->getId()==getCompiler()->fw_id ) {

	    list<FWObject*> l2=getCompiler()->fw->getByType(Interface::TYPENAME);
	    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
		Interface *interface_=Interface::cast(*i);
		if (interface_->isExt()) 
		    cl.push_back(interface_);
	    }
	    if ( ! cl.empty() ) {
		rel->clearChildren();

		for (FWObject::iterator i1=cl.begin(); i1!=cl.end(); ++i1) 
		{
		    rel->addRef( *i1 );
		}
	    }
	}
    }
}

/*
 *  I assume that there is always only one object in ODst, TSrc and TDst
 *  rule elements. This should have been assured by inspector VerifyRules
 */
void NATCompiler_pf::ReplaceObjectsTDst::operator()(NATRule *rule) const
{
    RuleElementTDst *rel;

    if (rule->getRuleType()==NATRule::Redirect) {
	rel=rule->getTDst();      assert(rel);
	rel->clearChildren();
	rel->addRef( getCompiler()->loopback_address );
    }
}

void NATCompiler_pf::compile()
{
    bool manage_virtual_addr=fwopt->getBool("manage_virtual_addr");


    cout << " Compiling NAT rules for " << fw->getName() << " ..." << endl << flush;

    try {

	Compiler::compile();

        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand groups ...\n" << flush;
	for_each_rule( ExpandGroups(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "split rule on original service ..\n" << flush;
	for_each_rule( splitOnOSrv(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "fill translated service ..\n" << flush;
	for_each_rule( fillTranslatedSrv(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "determine NAT rule types ..\n" << flush;
	for_each_rule( NATRuleType(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "verify NAT rules ..\n" << flush;
	for_each_rule( VerifyRules(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "replace references to the firewall ...\n" << flush;
	for_each_rule( ReplaceFirewallObjectsODst(this) );
	for_each_rule( ReplaceFirewallObjectsTSrc(this) );
	for_each_rule( ReplaceObjectsTDst(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

        if (verbose) cout << "process negation ...\n" << flush;
        for_each_rule( doOSrcNegation(this) );
        for_each_rule( doODstNegation(this) );
        for_each_rule( doOSrvNegation(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if ( manage_virtual_addr ) {
	    if (verbose) cout << "add virtual addresses for NAT rules ...\n" << flush;
	    for_each_rule( addVirtualAddress(this) );
            if (debug_rule>=0) for_each_rule( debugRule(this) );
	}

	if (verbose) cout << "expand multiple addresses ...\n" << flush;
	for_each_rule( ExpandMultipleAddresses(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "expand address range objects  ...\n" << flush;
	for_each_rule( addressRanges(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "convert to atomic rules ...\n" << flush;
	for_each_rule( ConvertToAtomic(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );

	if (verbose) cout << "assign rules to interfaces ...\n" << flush;
	for_each_rule( AssignInterface(this) );
        if (debug_rule>=0) for_each_rule( debugRule(this) );


    } catch (FWException &ex) {
	abort(ex.toString());
    }
}


void NATCompiler_pf::epilog()
{
    try {
	if (verbose) cout << "printing ... \n" << flush;

/* print access lists here */
	PrintAtomicRule par(this);

	string current_rule_label="";

	for (list<FWObject*>::iterator k=combined_ruleset->begin(); k!=combined_ruleset->end(); ++k) {
	    NATRule *r = NATRule::cast( *k );
	    string rl=r->getStr("rule_label");
	    if (rl!=current_rule_label) {

                if (verbose) {
                    cout << rl << endl << flush;
                }

		output << "# " << endl;
		output << "# Rule  " << rl << endl;

                string    comm=r->getComment();
                string::size_type c1,c2;
                c1=0;
                while ( (c2=comm.find('\n',c1))!=string::npos ) {
                    output << "# " << comm.substr(c1,c2-c1) << endl;
                    c1=c2+1;
                }
                output << "# " << comm.substr(c1) << endl;
                output << "# " << endl;

		current_rule_label=rl;
	    }
	    par(r);
	}

	cout << " NAT rules compiled successfully \n" << flush;

    } catch (FWException ex) {
	abort(ex.toString());
    }
}
