/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler_pf.hh,v 1.19 2002/04/18 22:11:02 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _POLICYCOMPILER_PF_HH
#define _POLICYCOMPILER_PF_HH

#include "fwcompiler/PolicyCompiler.hh"


namespace libfwbuilder {
class TCPService;
};

namespace fwcompiler {


    class PolicyCompiler_pf : public PolicyCompiler {

	public:

	/**
	 * set of methods for printing
	 */

	virtual void _printSrcService(libfwbuilder::RuleElementSrv  *o);
	virtual void _printDstService(libfwbuilder::RuleElementSrv  *o);
	virtual void _printProtocol(libfwbuilder::Service *srv);

	virtual std::string _printPort(int rs,int re,bool neg=false);
	virtual std::string _printSrcService(libfwbuilder::Service *srv,bool neg=false);
	virtual std::string _printDstService(libfwbuilder::Service *srv,bool neg=false);
	virtual std::string _printTCPFlags(libfwbuilder::TCPService *srv);

	virtual void _printSrcAddr(libfwbuilder::RuleElementSrc  *o);
	virtual void _printDstAddr(libfwbuilder::RuleElementDst  *o);
	virtual void _printAddr(libfwbuilder::Address  *o,bool neg=false);

	virtual void _printNegation(libfwbuilder::RuleElement  *o);

	virtual void _printAction(libfwbuilder::PolicyRule *r);
	virtual void _printLogging(libfwbuilder::PolicyRule *r);
	virtual void _printDirection(libfwbuilder::PolicyRule *r);
	virtual void _printInterface(libfwbuilder::PolicyRule *r);

	/**
	 * our firewall policy must block everything by default even
	 * if there are no rules. In iptables we do this by setting
	 * default chain policies to DROP. Here we do this by adding
	 * this unconditional blocking rule in the end. See also comment
	 * in the code regarding "pass_all_out" option
	 */
	void addDefaultPolicyRule();

	protected:

	/**
	 *  this rule processor splits rule if one of the objects in Src
	 *  is firewall itself. This is needed to properly choose direction 
	 *  later in filDirection
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, splitIfFirewallInSrc );


	/**
	 *  this rule processor splits rule if one of the objects in Dst
	 *  is firewall itself. This is needed to properly choose direction 
	 *  later in filDirection
	 */
        DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, splitIfFirewallInDst );


        /**
	 *  This inspector decides on direction if it is empty.
	 *
	 *  Algorithm is as follows:
	 *
	 *   I now support two modes for this compiler:
	 *
	 *   1. compiler produces two pf (or ipf) rules per each
	 *   global policy rule, one "Inbound" and another
	 *   "Outbound". Predicate SplitDirection does this for me if
	 *   I set direction to Both here.
	 *
	 *      Special cases:
	 *        If Src is single object which is firewall itself, then set
	 *        direction to Outbound
	 *        If Dst is single object which is firewall itself, then set
	 *        direction to Inbound
	 *        If oth Src and Dst contain firewall, then set directon to
	 *        Both
	 *
	 *      In fact predicates splitIfFirewallInSrc and
	 *      splitIfFirewallInDst make sure that if firewall is in
	 *      Src or Dst, then it is the only object there. Thus we
	 *      do not need to check number of objects in Src and Dst.
	 *
	 *   2. compiler produces one pf (or ipf) "in" rule per each
	 *   global policy rule. In this case I set direction to
	 *   Inbound here.
	 *      
	 *   I distinguish modes using firewall option "pass_all_out"
	 *
	 *             03/21/02   --vk
	 */
        DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_pf, fillDirection );

	/**
	 * split rules if direction is "Both"
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, SplitDirection );

	/**
	 * Option "scrub" does not accept "quick" and therefore does not
	 * stop matching sequence. We need to split rule onto two, the first
	 * will be generated with action "scrub", while the second one with
	 * action "pass" and option "quick"
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, ProcessScrubOption );

	/**
	 * this processor checks for the services which require
	 * special treatment.  Some of these will be checking for
	 * source or destination object as well because special
	 * command may need to be generated in case source or
	 * destination is a firewall itself. Therefore this processor
	 * should be called after converting to atomic rules, but
	 * before interface addresses in source and destination are
	 * expanded.
	 */
	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_pf, SpecialServices );

	/**
	 * PF does not support filtering by MAC addresses. Throw
	 * exception if such filtering is requested in the rule.
	 */
	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_pf, MACFiltering );

	/**
	 * this inspector sets 'quick' flag on rules
	 */
	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_pf, setQuickFlag );

	/**
	 * this processor deals with negation in Src in policy rules.
	 * THIS MUST BE THE LAST PROCESSOR
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, doSrcNegation );

	/**
	 * this processor deals with negation in Dst in policy rules.
	 * THIS MUST BE THE LAST PROCESSOR
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, doDstNegation );

	/**
	 * this processor deals with negation in Srv in policy rules.
	 * THIS MUST BE THE LAST PROCESSOR
	 *
	 * NOT IMPLEMENTED 
	 */
	DECLARE_POLICY_RULE_PROCESSOR( PolicyCompiler_pf, doSrvNegation );

	/**
	 *  this inspector prints single policy rule, assuming all
	 *  groups have been expanded, so source, destination and
	 *  service hold exactly one object each, and this object is
	 *  not a group.  Negation should also have been taken care of
	 *  before this method is called.
	 */
 	DECLARE_POLICY_RULE_INSPECTOR( PolicyCompiler_pf, PrintRule );

	virtual std::string myPlatformName();


	public:

	PolicyCompiler_pf(libfwbuilder::FWObjectDatabase *_db,
			  const std::string &fwname,
			  fwcompiler::OSConfigurator *_oscnf) : PolicyCompiler(_db,fwname,_oscnf) {}


	virtual int  prolog();
	virtual void compile();
	virtual void epilog();
	
    };


}

#endif
