/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: ipf.cc,v 1.16 2002/05/12 07:47:42 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

#include <fstream>
#include <algorithm>
#include <functional>
#include <iostream>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef __MINGW32__
#include <assert.h>
#endif

#include "PolicyCompiler_ipf.hh"
#include "NATCompiler_ipf.hh"

#include "OSConfigurator_openbsd.hh"
#include "OSConfigurator_freebsd.hh"
#include "OSConfigurator_solaris.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/XMLTools.hh"
#include "fwbuilder/FWException.hh"

#include "fwcompiler/manifest.hh"

#ifdef HAVE_GETOPT_H
  #include <getopt.h>
#else
  #ifdef __MINGW32__
    #include <getopt.h>
  #else
    #include <stdlib.h>
  #endif
#endif



static const char      *filename       = NULL;
static const char      *wdir           = NULL;
static const char      *mfile          = NULL;
static const char      *fwobjectname   = NULL;
static int              dl             = 0;
static int              drp            = -1;
static int              drn            = -1;
static int              verbose        = 0;

using namespace libfwbuilder;
using namespace fwcompiler;

class UpgradePredicate: public XMLTools::UpgradePredicate
{
    public:
    virtual bool operator()(const string &msg) const 
    { 
	cout << "Data file has been created in the old version of Firewall Builder. Use fwbuilder GUI to convert it." << endl;
	return false;
    }
};
    
void usage(const char *name)
{
    cout << "Firewall Builder:  policy compiler for FreeBSD ipf" << endl;
    cout << "Version " << VERSION << endl;
    cout << "Usage: " << name << " [-x] [-v] [-V] [-f filename.xml] [-d destdir] [-m manifest] firewall_object_name" << endl;
}

/**
 *  TODO:  move this to fwcompiler
 */
Firewall* findFirewallByName(FWObjectDatabase* db,const string &name)
{
    FWObject *fwgrp=db->getById( FWObjectDatabase::db->std.FirewallsId , true );
    assert(fwgrp!=NULL);

    FWObject::const_iterator i=
	std::find_if( fwgrp->begin(), fwgrp->end(), FWObjectNameEQPredicate(name) );

    return (i==fwgrp->end())?NULL:Firewall::cast( *i );
}

int main(int argc, char * const *argv)
{   

    if (argc<=1)
    {
        usage(argv[0]);
        exit(1);
    }

    int   opt;

    while( (opt=getopt(argc,argv,"x:vVf:d:m:")) != EOF )
    {
        switch(opt)
        {
        case 'd':
            wdir = strdup(optarg);
            break;
        case 'm':
            mfile = strdup(optarg);
            break;
        case 'f':
            filename = strdup(optarg);
            break;
        case 'x':
            if (*optarg=='r') {
                ++optarg;
                drp  = atoi(optarg);
            } else
                if (*optarg=='n') {
                    ++optarg;
                    drn  = atoi(optarg);
                } else
                    dl++;  // increase debug level
            break;
        case 'v':
            verbose++;
            break;
        case 'V':
            usage(argv[0]);
            exit(1);
        }
    }
    
    if((argc-1) != optind)
    {
        usage(argv[0]);
        exit(1);
    }

    fwobjectname = strdup( argv[optind++] );


    if (wdir==0) 	wdir="./";

    if(chdir(wdir))   {
	cerr << "Can't change to: " << wdir << endl;
	exit(1);
    }


    try   {

	/* init threads */
	g_thread_init(NULL);
    
        libfwbuilder::init();

	/* Initialize XML parser  */
	XMLTools::initXMLTools();

	/* create database */
	new FWObjectDatabase();

	/* load the data file */
	UpgradePredicate upgrade_predicate; 

	if (verbose) cout << " *** Loading data ...";
	FWObjectDatabase::db->load(filename,  &upgrade_predicate);
	if (verbose) cout << " done\n";


	string   fw_file_name=string(fwobjectname)+".fw";
	ofstream fw_file(fw_file_name.c_str());

	/* Review firewall and OS options and generate commands */
	Firewall*  fw=findFirewallByName(FWObjectDatabase::db, fwobjectname);
	FWOptions* options=fw->getOptionsObject();
	string s;

	string firewall_dir=options->getStr("firewall_dir");
	if (firewall_dir=="") firewall_dir="/etc";

	bool debug=options->getBool("debug");
	string shell_dbg=(debug)?"-x":"" ;
	string ipf_dbg=(debug)?"-v":"";

	fw_file << "#!/bin/sh " << shell_dbg << endl << endl;

	fw_file << "cd " << firewall_dir << " || exit 1" << endl << endl;

	fw_file << endl 
		<< "ipf -Fa" << endl
		<< "ipnat -C" << endl;

	fw_file << endl;

/*
 * Process firewall options, build OS network configuration script
 */
	OSConfigurator *oscnf=NULL;
	if (fw->getStr("host_OS")=="solaris")   oscnf=new OSConfigurator_solaris(FWObjectDatabase::db , fwobjectname);
	if (fw->getStr("host_OS")=="openbsd")   oscnf=new OSConfigurator_openbsd(FWObjectDatabase::db , fwobjectname);
	if (fw->getStr("host_OS")=="freebsd")   oscnf=new OSConfigurator_freebsd(FWObjectDatabase::db , fwobjectname);
	if (oscnf==NULL)
	    throw FWException("Unrecognized host OS "+fw->getStr("host_OS"));

	oscnf->processFirewallOptions();



/*
 * create compilers and run the whole thing 
 */
        Manifest manifest;
	Action a;


	PolicyCompiler_ipf c( FWObjectDatabase::db , fwobjectname , oscnf );

	c.setDebugLevel( dl );
	c.setDebugRule(  drp );
	c.setVerbose( verbose );

	string   ipf_file_name=string(fwobjectname)+"-ipf.conf";
	ofstream ipf_file(ipf_file_name.c_str());
	bool     have_ipf=false;
	if ( c.prolog() > 0 ) {

	    have_ipf=true;

	    c.compile();
	    c.epilog();

	    a.action     = "COPY";
	    a.filename   = ipf_file_name;
	    a.parameters = firewall_dir;
	    manifest.push_back(a);
	}


	NATCompiler_ipf n( FWObjectDatabase::db , fwobjectname , oscnf );

	n.setDebugLevel( dl );
	n.setDebugRule(  drn );
	n.setVerbose( verbose );

	string   nat_file_name=string(fwobjectname)+"-nat.conf";
	ofstream nat_file(nat_file_name.c_str());
	bool     have_nat;
	if ( n.prolog() > 0 ) {

	    have_nat=true;

	    n.compile();
	    n.epilog();

	    a.action     = "COPY";
	    a.filename   = nat_file_name;
	    a.parameters = firewall_dir;
	    manifest.push_back(a);
	}
/*
 * now write generated scripts to files
 */
	fw_file << oscnf->getCompiledScript();

	if (have_ipf) {
	    ipf_file << c.getCompiledScript();
	    ipf_file.close();
	    fw_file << "ipf " << ipf_dbg << " -f " << ipf_file_name << endl;
	}

	if (have_nat) {
	    nat_file << n.getCompiledScript();
	    nat_file.close();
	    fw_file << "ipnat " << ipf_dbg << " -f " << nat_file_name <<  endl;
	}

	fw_file << endl;

	fw_file << "ipf -E" << endl;

	fw_file << endl;

	if(mfile)
            manifest.save(string(mfile), S_IRUSR|S_IWUSR);
        
        return 0;

    } catch(FWException &ex)  {
	cerr << ex.toString() << endl;
        return 1;
    } catch (std::string s) {
	cerr << s;
        return 1;
    } catch (std::exception ex) {
	cerr << ex.what();
        return 1;
    } catch (...) {
	cerr << "Unsupported exception";
        return 1;
    }

}









